@@ -1,15 +1,16 @@
 import os
-import requests
 import csv
 import json
+import httpx
 from typing import Union
-from datetime import datetime
+from datetime import datetime, timedelta
 from .config import Frequency, AssetType, AssetUrl
-
+from .utils import VinterValidation, VinterUrl
+from .vinter_abc import VinterAPIABC
 
 APIKEY = os.environ.get("VINTER_API_KEY", None)
 
-class VinterAPI:
+class VinterAPI(VinterAPIABC):
     def __init__(self, api_key: str, asset_type: str):
         """This function takes in an api_key and asset_type and sets them as attributes of the class
 
@@ -18,17 +19,16 @@ class VinterAPI:
         api_key : str
             Your API key.
         asset_type : str
-            The type of asset you want to get data for. This can be one of the following:
-            ["multi_assets", "single_assets"]
+            The type of asset you want to get data for. The acceptable asset types listed in the AssetType enum.
         """
         self.api_key = api_key
         self.asset_type = asset_type
         self.frequencies = [frequency.value for frequency in Frequency]
         self.valid_asset_types = [asset_type.value for asset_type in AssetType]
-        self.valid_symbols = None
-        self.validate_asset_type()
+        VinterValidation.validate_asset_type(self.asset_type)
+        self.httpx_client = httpx.Client(follow_redirects=True, timeout=10)
 
-    def get_all_active_symbols(self) -> dict:
+    def get_all_active_symbols(self, frequency: str = None, symbol_only: bool = False) -> Union[list, dict]:
         """This function returns a dictionary of all the active symbols
 
         Returns
@@ -36,101 +36,25 @@ class VinterAPI:
             A dictionary of all the active symbols
 
         """
-        url = None
-        for asset_url in AssetUrl:
-            if (
-                asset_url.value["asset_type"].value == self.asset_type
-                and asset_url.value["frequency"] is None
-            ):
-                url = asset_url.value["url"]
-                break
-
-        if url is None:
-            raise ValueError(
-                f"The asset type must be in {self.valid_asset_types}"
-            )
-
+        url = VinterUrl.get_active_url(self.asset_type)
         headers = {}
-        response = requests.get(url, headers=headers)
+        response = self.httpx_client.get(url, headers=headers)
+
+        response.raise_for_status() # Raise an exception if the request failed
+
         data = response.json()["data"]
-        self.valid_symbols = [symbol["symbol"] for symbol in data]
+
+        if frequency is not None:
+
+            VinterValidation.validate_frequency(frequency)
+
+            data = [asset for asset in data if asset['symbol'].split('-')[-1] == frequency]
+
+        if symbol_only:
+            data = [asset["symbol"] for asset in data]
+
         return data
 
-    def populate_valid_symbols(self) -> None:
-        """This function populates the valid_symbols attribute of the class with a list of all the
-        active symbols
-
-        Returns
-        -------
-
-        """
-        self.valid_symbols = [symbol["symbol"] for symbol in self.get_all_active_symbols()]
-
-    def validate_asset_type(self) -> None:
-        """If the asset type is not a valid asset_type , then raise a
-        ValueError
-
-        """
-        if self.asset_type not in self.valid_asset_types:
-            raise ValueError(
-                f"The asset type must be one of the following : {self.valid_asset_types}"
-            )
-
-    def validate_symbol_frequency(self, symbol: str) -> str:
-        """This function takes in a symbol and returns the symbol with the frequency appended to it.
-
-        Parameters
-        ----------
-        symbol : str
-            The symbol of the asset you want to get data for.
-
-        Returns
-        -------
-            The symbol with the frequency appended to it.
-
-        """
-        sym_frequency = symbol.split("-")[-1]
-
-        if sym_frequency not in self.frequencies:
-            raise ValueError(
-                "The frequency must be one of the following valid frequencies: {}".format(
-                    self.frequencies
-                )
-            )
-
-        self.frequency = sym_frequency
-
-        return symbol, sym_frequency
-
-    def get_url_by_asset_type(self, symbol: str) -> str:
-        """This function takes in a symbol and returns the url to use to get the data
-
-        Parameters
-        ----------
-        symbol : str
-            The symbol of the asset you want to get data for.
-
-        Returns
-        -------
-            The url to use to get the data
-
-        """
-
-        symbol, frequency = self.validate_symbol_frequency(symbol)
-
-        url = None
-        for asset_url in AssetUrl:
-            if (
-                asset_url.value["asset_type"].value == self.asset_type
-                and asset_url.value["frequency"].value == frequency
-            ):
-                url = asset_url.value["url"]
-                break
-
-        if url is None:
-            raise ValueError("The URL could not be found.")
-
-        return url
 
     def get_latest_data(self, symbol: str, limit: int = 1) -> dict:
         """It takes a symbol and a limit as parameters, and returns a dictionary of the latest data for
@@ -148,25 +72,17 @@ class VinterAPI:
             A dictionary of the latest data for the symbol and limit.
 
         """
-        url = self.get_url_by_asset_type(symbol=symbol)
+        url = VinterUrl.get_url_by_symbol(self.asset_type, symbol)
 
         params = {"symbol": symbol, "limit": limit}
         headers = {"Authorization": self.api_key}
-        response = requests.get(url, params=params, headers=headers)
+        response = self.httpx_client.get(url, params=params, headers=headers)
 
         response.raise_for_status()
 
         data = response.json()["data"]
 
         if len(data) == 0:
-            if self.valid_symbols is None:
-                self.populate_valid_symbols()
-
-            if symbol not in self.valid_symbols:
-                raise ValueError(
-                    f"The symbol is not a present in the list of active symbols for asset_type {self.asset_type}."
-                )
-                
             raise ValueError("No data was found for the symbol: {}".format(symbol))
 
         return data
@@ -186,6 +102,23 @@ class VinterAPI:
         """
         data = self.get_latest_data(symbol=symbol)
         return data[0]["value"]
+    
+    def _filter_by_symbol(self, data: list, symbol: str) -> list:
+        """This function takes in a list of data and a symbol and returns a list of data for that symbol
+
+        Parameters
+        ----------
+        data : list
+            A list of data
+        symbol : str
+            The symbol of the asset you want to get data for.
+
+        Returns
+        -------
+            A list of data for the symbol
+
+        """
+        return [asset for asset in data if asset["symbol"] == symbol]
 
     def _get_active_asset_data(self, symbol: str) -> dict:
         """This function returns the data for the active asset
@@ -201,27 +134,16 @@ class VinterAPI:
 
         """
 
-        symbol, frequency = self.validate_symbol_frequency(symbol)
+        symbol, frequency = VinterValidation.validate_symbol_frequency(symbol)
         
         data = self.get_all_active_symbols()
 
-        output = {}
-
-        for asset in data:
-            if asset["symbol"] == symbol:
-                output = asset
-                break
+        output = self._filter_by_symbol(data=data, symbol=symbol)
 
         if len(output) == 0:
-
-            if symbol not in self.valid_symbols:
-                raise ValueError(
-                    f"The symbol is not a present in the list of active symbols for asset_type {self.asset_type}."
-                )
-
             raise ValueError("No data was found for the symbol: {}".format(symbol))
 
-        return output
+        return output[0]
 
     def get_current_rebalance_weight(self, symbol: str) -> dict:
         """This function returns the current rebalance weight of multi_assets symbol
@@ -245,12 +167,11 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "weights" in data:
-            output = data["weights"]
+        output = data.get("weights", None)
 
-        if output == "" or output is None:
+        if output is None or output == "":
             raise ValueError("No data was found for the symbol: {}".format(symbol))
-
+        
         return output
 
     def get_contributions(self, symbol: str) -> dict:
@@ -275,8 +196,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "contrib" in data:
-            output = data["contrib"]
+        output = data.get("contributions", None)
 
         if output == "" or output is None:
             raise ValueError(
@@ -311,8 +231,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "previous_rebalance_date" in data:
-            output = data["previous_rebalance_date"]
+        output = data.get("previous_rebalance_date", None)
 
         return output
 
@@ -342,8 +261,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "previous_review_date" in data:
-            output = data["previous_review_date"]
+        output = data.get("previous_review_date", None)
 
         return output
 
@@ -373,8 +291,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "next_review_date" in data:
-            output = data["next_review_date"]
+        output = data.get("next_review_date", None)
 
         return output
 
@@ -404,8 +321,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "next_rebalance_date" in data:
-            output = data["next_rebalance_date"]
+        output = data.get("next_rebalance_date", None)
 
         return output
 
@@ -421,12 +337,8 @@ class VinterAPI:
             ValueError if the symbol is not a present in the list of active symbols for asset_type multi_assets
             
             OR
-            
-            ValueError if the next rebalance date is in the future
 
-            OR
-
-            None if the symbol Rebalance is not scheduled
+            None if the symbol Rebalance is not present in the payload
 
         """
 
@@ -439,26 +351,7 @@ class VinterAPI:
 
         data = self._get_active_asset_data(symbol=symbol)
 
-        if "next_rebalance_weights" in data:
-            output = data["next_rebalance_weights"]
-
-        if output is None:
-
-            if data["next_rebalance_date"] is None:
-                raise ValueError(
-                    "The next rebalance date is not scheduled, therefore the next rebalance weight is not available"
-                )
-
-            reb_date = datetime.strptime(data["next_rebalance_date"], "%Y-%m-%d")
-
-            if reb_date > datetime.utcnow():
-                raise ValueError(
-                    "The next rebalance date is in the future, therefore the next rebalance weight is not available"
-                )
-            else:
-                raise ValueError(
-                    "The next rebalance weight is not available, please contact support"
-                )
+        output = data.get("next_rebalance_weights", None)
 
         return output
 
@@ -480,7 +373,7 @@ class VinterAPI:
             A dictionary of the data
 
         """
-        symbol, frequency = self.validate_symbol_frequency(symbol)
+        symbol, frequency = VinterValidation.validate_symbol_frequency(symbol)
 
         if frequency != Frequency.DAILY.value:
             raise ValueError("The frequency must be daily to use this function.")
@@ -489,49 +382,52 @@ class VinterAPI:
             dates = [dates]
 
         # Validate Dates with regex pattern & Date validation
-        self.validate_dates(dates)
+        VinterValidation.validate_dates(dates)
 
-        output = []
-        
-        data = self.get_latest_data(symbol=symbol, limit=2000)   
+        start_date, last_date = dates[0], dates[-1]
 
-        for date in dates:
-            date_data = []
-            for candle in data:
-                candle_date = candle["created_at"].split("T")[0]
-                if candle_date == date:
-                    date_data.append(candle)
+        # Adding 1 day to the last date to get the data for the last date
+        last_date = datetime.strptime(last_date, "%Y-%m-%d") + timedelta(days=1)
 
-            if len(date_data) == 0:
-                print(
-                    "No data was found for the symbol: {} and date: {}".format(
-                        symbol, date
-                    )
-                )
+        # Converting the datetime object to string
+        last_date = last_date.strftime("%Y-%m-%d")
 
-            output.extend(date_data)
+        data = self.get_data_by_time(symbol=symbol, start=start_date, end=last_date)
 
-        return output
-
-    def validate_dates(self, dates: list) -> None:
-        """It raises a ValueError if any of the dates in the dates list are not in the format YYYY-MM-DD
+        return data
+           
+    def get_data_by_time(self, symbol: str, start: str, end: str = None, limit: int = 1000) -> dict:
+        """This function takes in a symbol and a start and end date and returns a dictionary of the data
+        for that period
 
         Parameters
         ----------
-        dates
-            A list of dates in the format YYYY-MM-DD
+        symbol : str
+            The symbol of the asset you want to get data for.
+        start : str
+            The start datatime . format: YYYY-MM-DDTHH:MM:SSZ or YYYY-MM-DDTHH:MM:SS.sssZ
+        end : str
+            The end datatime. format: YYYY-MM-DDTHH:MM:SSZ or YYYY-MM-DDTHH:MM:SS.sssZ
+
+        Returns
+        -------
+            A dictionary of the data
 
         """
-        for date in dates:
-            # Validate that the date is a valid date
-            try:
-                datetime.strptime(date, "%Y-%m-%d")
-            except ValueError:
-                raise ValueError(
-                    "The date must be in the format YYYY-MM-DD. The date: {} is not in the correct format.".format(
-                        date
-                    )
-                )
+        url = VinterUrl.get_url_by_symbol(asset_type=self.asset_type, symbol=symbol)
+
+        params = {"symbol": symbol, "start_time": start, "end_time": end, "limit": limit}
+        headers = {"Authorization": self.api_key}
+        response = self.httpx_client.get(url, params=params, headers=headers)
+
+        response.raise_for_status() # Raise an exception if the request fails
+
+        data = response.json()["data"]
+
+        if len(data) == 0:               
+            raise ValueError(f"No data was found for the symbol: {symbol} between {start} and {end}.")
+
+        return data
 
     def save_data_to_file(
         self, data: dict, filename: str, file_type: str = "csv", seprator: str = ","
